use std::fmt;
use std::error::Error as StdError;
use std::ops::Deref;
use futures::{future, Future, Stream};
use futures::future::BoxFuture;
use hyper::mime::{self, Mime};
use url::form_urlencoded::{self, Parse};
use errors::HandleError;
use super::body::Body;
use super::from_body::FromBody;


pub trait FromForm: Sized {
    type Err;
    fn from_form<'a>(iter: Parse<'a>) -> Result<Self, Self::Err>;
}


pub struct Form<F: FromForm>(F);

impl<F: FromForm> Deref for Form<F> {
    type Target = F;
    fn deref(&self) -> &Self::Target {
        &self.0
    }
}

impl<F> FromBody for Form<F>
where
    F: Send + 'static + FromForm,
{
    type Future = BoxFuture<Form<F>, HandleError>;

    fn check_mime(m: &Mime) -> bool {
        *m == mime::APPLICATION_WWW_FORM_URLENCODED
    }

    fn from_body(body: Body) -> Self::Future {
        body.fold(Vec::new(), super::fold_fn)
            .and_then(|body| {
                let iter = form_urlencoded::parse(&body);
                future::result(F::from_form(iter).map(Form).map_err(|_| {
                    HandleError::bad_request()
                }))
            })
            .boxed()
    }
}


#[doc(hidden)]
#[derive(Debug)]
pub enum FormParseError {
    InvalidKey,
    MissingKey,
    DuplicatedKey,
}

impl fmt::Display for FormParseError {
    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
        write!(f, "{:?}", self)
    }
}

impl StdError for FormParseError {
    fn description(&self) -> &str {
        "form parse error"
    }
}
