use std::fmt;
use std::error::Error as StdError;
use hyper::Request as HyperRequest;
use hyper::header::ContentType;
use futures::{future, Future, BoxFuture};
use typemap::SendMap;
use tokio_core::reactor::Remote;

use body::{Body, FromBody};
use errors::HandleError;
use outcome::Outcome;
use request::{self, Request};
use response::{Response, Responder, Callback};
use router::regex::OwnedCaptures;


/// A context during handling.
///
/// It contains an HTTP request, HTTP response to return,
/// and a typemap in order to share variables between middlewares.
pub struct Context {
    /// Incoming request contains URI, method, headers, and body
    pub req: Request,
    /// Captured parameters from router
    pub params: OwnedCaptures,
    /// Request body
    pub(crate) body: Option<Body>,
    /// Handle of event loop
    pub handle: Remote,
    /// Extension map to share values between handlers/middlewares
    pub ext: SendMap,
    pub(crate) on_send: Vec<Callback>,
}

impl Context {
    pub(crate) fn new(req: HyperRequest, params: OwnedCaptures, handle: Remote) -> Self {
        let (req, body) = request::reconstruct(req);
        Context {
            req,
            params,
            body: Some(body.into()),
            handle,
            ext: SendMap::custom(),
            on_send: Vec::new(),
        }
    }

    pub fn respond<R: Responder>(self, responder: R) -> Outcome<Response> {
        match responder.respond(&self.req) {
            Ok(raw) => {
                let mut res = Response::from(raw);
                res.apply_callbacks(self.on_send);
                Outcome::continue_(res)
            }
            Err(err) => Outcome::halt(err),
        }
    }

    pub fn next(self) -> Outcome<Context> {
        Outcome::continue_(self)
    }

    pub fn halt<T>(self, err: HandleError) -> Outcome<T> {
        Outcome::halt(err)
    }

    /// Register a callback function called after construct response.
    pub fn on_send(&mut self, on_send: Callback) -> &mut Self {
        self.on_send.push(on_send);
        self
    }

    pub fn take_body<T>(&mut self) -> BoxFuture<T, HandleError>
    where
        T: Send + 'static + FromBody,
    {
        match self.req.headers.get::<ContentType>() {
            Some(&ContentType(ref m)) if T::check_mime(m) => (),
            _ => return future::err(HandleError::bad_request()).boxed(),
        }
        match self.body.take() {
            Some(body) => body.into_value::<T>().boxed(),
            None => future::err(HandleError::internal(EmptyBody)).boxed(),
        }
    }
}



#[derive(Debug)]
struct EmptyBody;

impl fmt::Display for EmptyBody {
    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
        write!(f, "{:?}", self)
    }
}

impl StdError for EmptyBody {
    fn description(&self) -> &str {
        "empty body"
    }
}
