use std::borrow::Cow;
use std::sync::Arc;
use futures::Future;
use hyper::Method;
use regex::Regex;

use context::Context;
use handler::{Handler, Middleware};
use outcome::Outcome;
use response::Response;
use super::regex::RegexPattern;


/// A route endpoint, which contains the instance of handler and middlewares
#[allow(dead_code)]
pub struct Route {
    pub(super) method: Method,
    pub(super) pattern: RegexPattern,
    pub(crate) middlewares: Vec<Arc<Middleware>>,
    pub(crate) handler: Arc<Handler>,
}

impl Route {
    pub fn new<S, H>(method: Method, pattern: S, handler: H) -> Self
    where
        S: AsRef<str>,
        H: Handler,
    {
        let pattern = normalize_pattern(pattern.as_ref());
        let pattern = Regex::new(&pattern).unwrap();
        Route {
            method: method,
            pattern: pattern.into(),
            middlewares: Vec::new(),
            handler: Arc::new(handler),
        }
    }

    pub fn with_middleware<M: Middleware>(mut self, middleware: M) -> Self {
        self.middlewares.push(Arc::new(middleware));
        self
    }

    #[inline(always)]
    pub fn get<S: AsRef<str>, H: Handler>(pattern: S, handler: H) -> Self {
        Self::new(Method::Get, pattern, handler)
    }

    #[inline(always)]
    pub fn post<S: AsRef<str>, H: Handler>(pattern: S, handler: H) -> Self {
        Self::new(Method::Post, pattern, handler)
    }

    #[inline(always)]
    pub fn put<S: AsRef<str>, H: Handler>(pattern: S, handler: H) -> Self {
        Self::new(Method::Put, pattern, handler)
    }

    #[inline(always)]
    pub fn delete<S: AsRef<str>, H: Handler>(pattern: S, handler: H) -> Self {
        Self::new(Method::Delete, pattern, handler)
    }

    #[inline(always)]
    pub fn head<S: AsRef<str>, H: Handler>(pattern: S, handler: H) -> Self {
        Self::new(Method::Head, pattern, handler)
    }

    pub(crate) fn apply<'a, I>(&'a self, ctx: Context, base_middlewares: I) -> Outcome<Response>
    where
        I: Iterator<Item = &'a Arc<Middleware>>,
    {
        let middlewares = base_middlewares.chain(self.middlewares.iter());
        let ctx = middlewares.fold(Outcome::continue_(ctx), |ctx, middleware| {
            let middleware = middleware.clone();
            Outcome::from_future(ctx.and_then(move |ctx| middleware.call(ctx)))
        });

        let handler = self.handler.clone();
        let outcome = ctx.and_then(move |ctx| handler.call(ctx));

        Outcome::from_future(outcome)
    }
}



fn normalize_pattern(pattern: &str) -> Cow<str> {
    let pattern = pattern
        .trim()
        .trim_left_matches("^")
        .trim_right_matches("$")
        .trim_right_matches("/");
    match pattern {
        "" => "^/$".into(),
        s => format!("^{}/?$", s).into(),
    }
}

#[cfg(test)]
mod tests {
    use super::normalize_pattern;

    #[test]
    fn normalize_cases() {
        assert_eq!(normalize_pattern("/"), "^/$");
        assert_eq!(normalize_pattern("/path/to"), "^/path/to/?$");
        assert_eq!(normalize_pattern("/path/to/"), "^/path/to/?$");
    }
}
