use std::collections::HashMap;
use std::sync::Arc;
use hyper::{Request as HyperRequest, StatusCode, Method};
use tokio_core::reactor::Remote;

use context::Context;
use errors::HandleError;
use handler::Middleware;
use outcome::Outcome;
use response::Response;
use super::regex::OwnedCaptures;
use super::route::Route;


#[derive(Default)]
pub struct Router {
    routes: HashMap<Method, Vec<Route>>,
}

impl Router {
    pub(crate) fn add_route(&mut self, route: Route) -> &mut Self {
        self.routes
            .entry(route.method.clone())
            .or_insert(Vec::new())
            .push(route);
        self
    }

    fn recognize(&self, method: &Method, path: &str) -> Option<(&Route, OwnedCaptures)> {
        self.routes.get(method).and_then(|routes| {
            for route in routes {
                if let Some(caps) = route.pattern.owned_captures(path) {
                    return Some((route, caps));
                }
            }
            None
        })
    }

    pub fn apply<'a, I>(
        &'a self,
        req: HyperRequest,
        handle: Remote,
        base_middlewares: I,
    ) -> Outcome<Response>
    where
        I: Iterator<Item = &'a Arc<Middleware>>,
    {
        match self.recognize(req.method(), req.path()) {
            Some((route, cap)) => {
                let ctx = Context::new(req, cap, handle);
                route.apply(ctx, base_middlewares)
            }
            None => Outcome::halt(HandleError::Status(StatusCode::NotFound)),
        }
    }
}
